#!BPY

""" Registration info for Blender menus:
Name: 'Chicken R44_nf (.egg)...'
Blender: 245
Group: 'Export'
Tip: 'Export to the Panda3D Egg format.'
"""
__author__ = "Daniel Amthauer, Simon Groenewolt"
__url__ = ("Chicken homepage, http://chicken-export.sf.net")
__email__="daniel*amthauer:gmail*com"
__version__ = "R44_nf"

__bpydoc__ = """
Chicken is an Egg exporter for Blender.<br>
For more information see the included HTML documentation, accessible through the
help button in the script's window.
"""

# chicken_exportR44_nf.py
# Copyright (C) 2008  Daniel Amthauer, Simon Groenewolt
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

import Blender, os, sys

from Blender import Draw, BGL, Image, Modifier, Mesh, Types, Window, Text
from Blender import Material as BMaterial
from Blender import Texture as BTexture
from Blender.Mathutils import *
from Blender.Window import DrawProgressBar, Theme
from Blender.sys import makename

webbrowser = None
INTERFACE_ERROR = False
try:
	import chicken_interface
except ImportError:
	exc = sys.exc_info()[1]
	msg = exc.args[0]
	err_str = 'chicken_interface could not be imported. Chicken will not work without this module.'
	err_str2 = 'The reported error was: '+msg
	if Blender.mode == 'background':
		print err_str
		print err_str2
	else:
		Draw.PupMenu('Chicken Error%s|%s|%s'%('%t', err_str, err_str2))
	INTERFACE_ERROR = True

if not INTERFACE_ERROR:
	from chicken_interface import *
	PATH = LoadPath()
	PATH_ERROR = CheckPath(PATH)

CHICKEN_LOGO = '/chicken/newchicken.png'
CHICKEN_DOC = '/chicken/doc.html'
MATERIALS, TEXTURES = None, None # globals for materials and textures
IDENTITY = Matrix() # 4x4 Identity matrix. Don't know if it's cheaper than making a new one every time, but I find it more readable
INITIAL = ["<CoordinateSystem> { Z-up }\n",
			'<Comment> { "Egg laid by Chicken for Blender v%s" }\n'%__version__]

FORCE_BACKGROUND = False

# optional features (depend on a recent CVS version of Blender, or experimental patches)
# set to True if you wish to test them
CURVE_SUPPORT = False # requires a Blender patch I made to support exporting curves. It was recently approved but it'll be a while until it's in an official release

class Exporter:
	'''Abstract base class for Gui and Background exporters'''
	def init(self, settings = None, progress_range=None):
		'''Not a real constructor, just a place to put common intialization'''
		self.messages = []
		self.meshes = {'static':[],
					   'armature':[]}
		self.error = False
		self.statusMsg = ''
		self.selection = None
		self.binPath = PATH
		# interpret progress range if any
		if type(progress_range) == type(str()):
			progress_range = map(float, progress_range.split(','))
			if len(progress_range) != 2:
				progress_range = (0.0, 1.0)
		elif progress_range is None:
			progress_range = (0.0, 1.0)
		self.prog_start = progress_range[0]
		self.prog_diff = progress_range[1]-progress_range[0]
		self.prog_end = progress_range[1]
		# begin load settings
		sDefault = settings is None # no user defined file or text to look in, therefore check defaults
		setting_txt = None
		if settings:
			settings_file = makename(settings, ext='.chicken')
		else:
			settings_file = makename(ext='.chicken')
		if os.path.isfile(settings_file):
			settings_file = open(settings_file, 'r')
			setting_txt = settings_file.read()
		else:
			if sDefault:
				settings = 'export.chicken'
			try:
				settings = Text.Get(settings)
			except NameError: pass
			else:
				setting_txt = '\n'.join(settings.asLines())

		if not setting_txt:
			print "Settings for export not found. Using defaults"
		# default values for possible settings
		d = {'filename': None,
			  'sel':None,
			  'anims':[],
			  'tools':(False,False,False),
			  'optparms':'',
			  'mextract':False,
			  'animonly':False,
                          'singlefile':False,
                          'dotangents':True
			  }
		if setting_txt:
			try:
				exec setting_txt in d
			except:
				print "Settings not valid. Exiting"
				return

		if d['filename']:
			self.filename = makename(d['filename'], ext='.egg')
		else:
			self.filename = makename(ext='.egg')
		self.anims = d['anims']
		o = Blender.Scene.GetCurrent().objects
		sel = d['sel']
		if sel:
			sel = set(sel)
			self.selection = filter(lambda x: x.name in sel, o)
		else:
			self.selection = filter(lambda x: x.sel, o)
		self.pview, self.egg2bam, self.optchar = d['tools']
		self.optParms = d['optparms']
		self.mExtract = d['mextract']
		self.animOnly = d['animonly']
		self.singleFile = d['singlefile']
		self.doTangents = d['dotangents']
		# end load settings

		self.success = False
		self.time = None
		self.normalMappedUVs = set([])
	def analyzeSelection(self):
		'''
		This method is called when updating the selection before exporting.
		It sets/updates variables that will be used upon export.
		'''
		self.success = False
		self.messages = []
		if self.selection:
			selection = self.selection
		else:
			selection = Blender.Object.GetSelected()
		self.lastSelected = selection
		# get all objects that are meshes, empties or curves
		objects = filter(lambda x: x.getType() in ('Mesh', 'Empty', 'Curve'), selection)
		if len(objects) == 0:
			self.messages.append(Message.no_selection)
		else:
			meshes = filter(lambda x: x.getType() == 'Mesh', objects)
			empties = filter(lambda x: x.getType() == 'Empty', objects)

			self.invalidModifiers = []
			self.envelopesUsed = []
			def getArmatureName(x):
				armatureMod = None
				for mod in x.modifiers:
					if mod.type == Modifier.Type.ARMATURE:
						armatureMod = mod
					else:
						self.invalidModifiers.append(x.name)
				if armatureMod:
					arm = armatureMod[Modifier.Settings.OBJECT]
					envelope = armatureMod[Modifier.Settings.ENVELOPES]
					if envelope:
						self.envelopesUsed.append(x.name)
					if arm:
						return arm.name
					else:
						return None
				return None
			armatures = map(getArmatureName, meshes) # get armature name for every selected mesh
			if self.envelopesUsed:
				self.messages.append(Message(Message.envelope_str%(', '.join(self.envelopesUsed)), Message.Warning))
			del self.envelopesUsed
			armatureSet = set(armatures) # set of unique armature names, None is in it if there are meshes with no armature
			armatureSet.discard(None)
			self.numArmatures = len(armatureSet)
			if self.numArmatures > 1:
				self.messages.append(Message.multi_armature)
				self.anims = None
			try:armature = armatureSet.pop()
			except KeyError: armature = None

			# divide meshes into static and animated (for now only armature)
			static_meshes, armature_meshes = [], []
			self.invalidModifiers = set(self.invalidModifiers)
			anim_mod_names = []
			for i, m in enumerate(meshes):
				if armatures[i]: 
					armature_meshes.append(m)
					if m.name in self.invalidModifiers:
						anim_mod_names.append(m.name)
				else: static_meshes.append(m)
			del self.invalidModifiers
			if anim_mod_names:
				self.messages.append(Message(Message.anim_mod_str%(', '.join(anim_mod_names)), Message.Warning))

			static_meshes.extend(empties) #put empties together with static meshes

			self.meshes = {'static':static_meshes,
						   'armature':armature_meshes}
			if armature:
				self.armature = Blender.Object.Get(armature)
				try:
					self.srb = self.armature.getData().bones['SyntheticRootBone']
				except KeyError: self.srb = None

			if CURVE_SUPPORT:
				curves = filter(lambda x: x.getType() == 'Curve', objects)
				self.curves = []
				for c in curves:
					curve = c.getData()
					for i, curnurb in enumerate(curve):
						if curnurb.type == 4 and not curnurb.isCyclic():
							self.curves.append((c.name+str(i), curnurb, c))

		self.analyzeMessages()

	def doExport(self):
		'''
		This method does the bulk of the export process. It creates the
		internal data structures to represent what will be exported and
		accumulates their representations in a buffer which is then
		written to a file. It then invokes any additional Panda tools that
		were specified by the user via the GUI or Background interfaces.
		'''
		self.filename = makename(self.filename, ext = ".egg")
		self.success = False
		self.time = None
		time1 = Blender.sys.time()
		if self.error:
			return
		self.messages = []
		self.reportProgress(self.prog_start, 'Exporting.')
		objs = {}
		global MATERIALS, TEXTURES
		MATERIALS, TEXTURES = {}, {}
		character = None
		if len(self.meshes['armature']) > 0:
			character = Character(self.armature, self.meshes['armature'], self.mExtract and self.srb, self.animOnly, addTangents=self.doTangents)
		if not self.animOnly:
			statics = self.meshes['static']
			if len(statics) > 0:
				groups = [Group(mesh=m, addTangents=self.doTangents) for m in statics]
				# begin reconstruction of hierarchy
				def getParentIndex(x):
					try: return statics.index(x.parent)
					except ValueError: return -1
				parents = map(getParentIndex, statics)
				for i, p in enumerate(parents):
					if p != -1:
						# if this group's parent is also in the selection
						# switch its transform to local and append it to the
						# parent group's children
						child = groups[i]
						child.transform = statics[i].matrixLocal
						groups[p].children.append(child)
					else:
						objs[groups[i].name] = groups[i]
				# end
			buffer = INITIAL[:]
			for mat in MATERIALS.values():
				if mat.users > 0:
					mat.write(buffer)
					#print mat
			for tex in TEXTURES.values():
				if tex.users > 0:
					tex.write(buffer)
					#print tex
			if character:
				character.write(buffer)
				#print character
			for o in objs.values():
				o.write(buffer)
				#print o
			if CURVE_SUPPORT:
				for c in self.curves:
					NurbsCurve(*c).write(buffer)
			if character: self.reportProgress(self.prog_start+0.5*self.prog_diff, 'Exporting..')
			else: self.reportProgress(self.prog_start+0.75*self.prog_diff, 'Exporting...')
			try:
				out_file = open(self.filename,"w")
			except IOError:
				self.messages.append(Message.file_error)
				self.analyzeMessages()
				self.statusMsg += ' ...Could not export'
				self.reportProgress(1.0, '')
				return
			else:
                                out_file.write('\n'.join(buffer))
                                if self.singleFile and character:
                                        self.reportProgress(self.prog_start+0.75*self.prog_diff, 'Exporting...')
					character.writeAnims("", self.anims, out_file)
				out_file.close()
		if character and not self.singleFile:
			self.reportProgress(self.prog_start+0.75*self.prog_diff, 'Exporting...')
			character.writeAnims(makename(self.filename, ext=''), self.anims)

		#begin calling of external Panda utils
		if self.egg2bam or self.optchar or self.pview or (len(self.normalMappedUVs) > 0 and self.doTangents==False):
			plainName = makename(self.filename, ext='')
			bamName= makename(self.filename,".bam")
			if self.anims:
				animNames = map(lambda x: "%s-%s"%(plainName,x[0]), self.anims)
			else: animNames = []
			argsPview = ['pview','-c', '-l']
			argsOpt = ['egg-optchar']
			argsTrans = ['egg-trans']
			if self.optParms:
				argsOpt.extend(self.optParms.split(' '))
			else:
				argsOpt.append('-inplace')

			if self.egg2bam:
				pviewExt = '.bam'
			else:
				pviewExt = '.egg'
			if os.name != 'nt': #in systems that have spawnvp use it
				spawnfunc = os.spawnvp
				argsBam = ['egg2bam', '-o'+bamName, self.filename]
				argsOpt.append(self.filename)
				if animNames:
					argsPview.append('-a')
				argsPview.append(makename(self.filename, ext=pviewExt))
				if animNames:
					argsPview.append(makename(animNames[0], ext=pviewExt))
				for n in animNames:
					argsOpt.append(makename(n,ext='.egg'))
				argsTrans.extend(['-o'+self.filename, self.filename])
			else:
				spawnfunc = os.spawnv
				argsBam = ['egg2bam', '-o"%s"'%bamName, '"%s"'%self.filename]
				argsOpt.append('"%s"'%self.filename)
				if animNames:
					argsPview.append('-i -a')
				argsPview.append('"%s"'%makename(plainName,ext=pviewExt))
				if animNames:
					argsPview.append('"%s"'%makename(animNames[0], ext=pviewExt))
				for n in animNames:
					argsOpt.append('"%s"'%makename(n,ext='.egg'))
				argsTrans.extend(['-o"%s"'%self.filename, '"%s"'%self.filename])

			if len(self.normalMappedUVs) > 0 and self.doTangents==False:
				argsTransUV = ' '.join(['-tbn %s'%name for name in self.normalMappedUVs])
				argsTrans[1:1] = argsTransUV.split(' ')
				try:
					call = os.path.join(self.binPath,'egg-trans')
					ret = spawnfunc(os.P_WAIT, call, argsTrans)
				except OSError:
						self.messages.append(Message.trans_nf)
				else:
					if ret != 0:
						self.messages.append(Message.trans_exec) 
			if self.optchar:
				try:
					call = os.path.join(self.binPath,'egg-optchar')
					ret = spawnfunc(os.P_WAIT, call, argsOpt)
				except OSError:
						self.messages.append(Message.opt_nf)
				else:
					if ret != 0:
						self.messages.append(Message.opt_exec)
			if self.egg2bam:
				if not self.animOnly:
					try:
						call = os.path.join(self.binPath,'egg2bam')
						ret = spawnfunc(os.P_WAIT, call, argsBam)
						print 'performed call %s with return code %d'%(call, ret)
					except OSError:
						self.messages.append(Message.bam_nf)
					else:
						if ret != 0:
							self.messages.append(Message.bam_exec)
				for n in animNames:
					if os.name == 'nt':
						argsBam[1] = '-o"%s"'%makename(n, ext='.bam')
						argsBam[2] = '"%s"'%makename(n, ext='.egg')
					else:
						argsBam[1] = '-o'+makename(n, ext='.bam')
						argsBam[2] = makename(n, ext='.egg')
					try:
						call = os.path.join(self.binPath,'egg2bam')
						ret = spawnfunc(os.P_WAIT, call, argsBam)
						print 'performed call %s with return code %d'%(call, ret)
					except OSError:
						self.messages.append(Message.bam_nf)
					else:
						if ret != 0:
							self.messages.append(Message.bam_exec)
			if self.pview:
				try:
					call = os.path.join(self.binPath,'pview')
					spawnfunc(os.P_NOWAIT, call, argsPview)
				except OSError:
					self.messages.append(Message.pv_nf)
		self.reportProgress(self.prog_end, 'Done')
		self.success = True
		self.time = Blender.sys.time() - time1
		self.analyzeMessages()

	def analyzeMessages(self):
		errors = 0
		warnings = 0
		for m in self.messages:
			if m.severity == Message.Error:
				errors += 1
			elif m.severity == Message.Warning:
				warnings += 1
		self.statusMsg = 'No apparent problems'
		self.error = errors > 0
		if errors or warnings:
			self.statusMsg = '%i Error(s), %i Warning(s) Found'%(errors,warnings)


class Gui(Exporter):
	'''
	This class encapsulates everything that involves interacting with
	the user through the Blender GUI
	'''
	animStart = 15 # the first event for animation delete buttons
	class AnimationRow:
		def __init__(self, name, fps, start, end):
			self.bName = ButtonWrapper('name: ', 0, 'string', 'The exported animation file will be named after this and the export file, so keep them unique', name, length=80)
			self.bFps = ButtonWrapper('fps:', 0, 'number', initial=fps, min=1, max=999)
			self.bStart = ButtonWrapper('start:', 0, 'number', initial=start, min=1, max=100000)
			self.bEnd = ButtonWrapper('end:', 0, 'number', initial=end, min=1, max=100000)
			self.bDelete = ButtonWrapper('Delete', 0, 'push')
		def update(self, i, w, wPad, top):
			self.bFps.update([wPad+5, top-1, 70, 20])
			self.bStart.update([wPad+75, top-1, 70, 20])
			self.bEnd.update([wPad+145, top-1, 70, 20])
			self.bName.update([wPad+215, top-1, w-2*wPad-270, 20])
			self.bDelete.evt = Gui.animStart+i
			self.bDelete.update([w-wPad-55, top-1, 50, 20])
		def getAnim(self):
			return [self.bName.val, self.bFps.val, self.bStart.val, self.bEnd.val]

	def __init__(self):
		self.init()
		self.selection = None
		self.reportProgress = DrawProgressBar
		self.page = 0
		self.vMsg = 'Version %s'%__version__
		self.vMsgW = Draw.GetStringWidth(self.vMsg)+5
		self.logo = None
		try: # try to load the logo from the datadir
			self.logo = Image.Load(Blender.Get("datadir")+CHICKEN_LOGO)
		except IOError:
			path = Blender.Get('udatadir')
			if path:
				try: # if that doesn't work, try the user data dir
					self.logo = Image.Load(path+CHICKEN_LOGO)
				except IOError: pass # give up

		self.buttonFuncs = {}

		self.bReport = ButtonWrapper('View Report Page', 9, 'push')
		def fReport(): self.page = (self.page+1)%2
		self.buttonFuncs[9] = fReport

		self.bUpdate = ButtonWrapper('Update Selection', 1, 'push', 'Re-scan Selected Objects')
		def analyzeAndLoad():
                        self.analyzeSelection()
                        self.loadState()
		self.buttonFuncs[1] = analyzeAndLoad

		self.bFile = ButtonWrapper('Export File: ', 0, 'string', initial=self.filename, length=300)
		self.bSFile = ButtonWrapper('Select', 3, 'push')
		def fSFile():
			def updateFile(fn):
				self.bFile.val = makename(fn, ext='.egg')
			Window.FileSelector(updateFile,"Select Export File",self.bFile.val)
		self.buttonFuncs[3] = fSFile

		self.bAddAnim = ButtonWrapper('Add Animation', 4, 'push')
		def fAddAnim():
			i = len(self.animationRows)+1
			rate = 24
			base = 0
			if len(self.animationRows)!=0:
                                rate = self.animationRows[-1].bFps.val
                                base = self.animationRows[-1].bEnd.val
			self.animationRows.append(Gui.AnimationRow('anim%i'%i, rate, base+1, base+rate))
		self.buttonFuncs[4] = fAddAnim

		self.bMExtraction = ButtonWrapper('Motion Extraction', 0, 'toggle', "The animation will be exported as if the SRB were the origin, and the displacements will be stored as the SRB's animation", self.mExtract)
		self.bAnimOnly = ButtonWrapper('Animation Only', 0, 'toggle', initial=self.animOnly)
		self.bSingleFile = ButtonWrapper('Single File', 0, 'toggle', initial=self.singleFile)
		self.bDoTangents = ButtonWrapper('Tangents & Binormals', 0, 'toggle', initial=self.doTangents)
		self.bQuit = ButtonWrapper('Quit', 6, 'push', 'Exit Chicken')
		self.buttonFuncs[6] = Draw.Exit

		self.bHelp = ButtonWrapper('Help', 7, 'push', 'See the documentation for the script (opens in web browser)')
		def fHelp():
			try: # try to find documentation in data dir
				path = 'file:///'+Blender.Get("datadir").replace('\\', '/')+CHICKEN_DOC
				webbrowser.open(path)
			except OSError:
				try: # if not, try in user data dir
					path = 'file:///'+Blender.Get("udatadir").replace('\\', '/')+CHICKEN_DOC
					webbrowser.open(path)
				except:
					Draw.PupMenu("Error%t|Documentation couldn't be found") # give up
		self.buttonFuncs[7] = fHelp

		self.bEgg2Bam = ButtonWrapper('egg2bam', 0, 'toggle', 'Export additional BAM files for all exported Eggs', self.egg2bam)
		self.bPview = ButtonWrapper('pview', 0, 'toggle', 'Preview exported Eggs. Only first animation will be shown.', self.pview)
		self.bEggOptchar = ButtonWrapper('egg-optchar', 0, 'toggle', 'Optimize an exported character and its animations', self.optchar)
		self.bOptParms = ButtonWrapper('Params:', 0, 'string', 'Parameters for optimization. If blank "-inplace" will be used.', self.optParms, length=80)
		self.bExport = ButtonWrapper('Export', 2, 'push', 'Export with current settings')
		def fExport():
			if self.error:
				Draw.PupMenu('There are errors that make exporting impossible|See the report page for details')
			else:
                                self.saveState()
                                self.singleFile = self.bSingleFile.val
				self.animOnly = self.bAnimOnly.val
				self.doTangents = self.bDoTangents.val
				self.egg2bam = self.bEgg2Bam.val
				self.pview = self.bPview.val
				self.optchar = self.bEggOptchar.val
				self.optParms = self.bOptParms.val
				self.filename = self.bFile.val
				self.mExtract = self.bMExtraction.val
				self.anims = [ar.getAnim() for ar in self.animationRows]
				self.doExport()
		self.buttonFuncs[2] = fExport

		self.bSaveSettings = ButtonWrapper('Save', 8, 'push', 'Save export settings')
		def fSaveSettings():
			txt = None
			try:
				txt = Text.Get('export.chicken')
			except NameError: pass
			else:
				txt.setName('export.chicken.old')
			txt = Text.New('export.chicken')
			selection = map(lambda x: x.name, self.lastSelected)
			tools = (self.bPview.val, self.bEgg2Bam.val, self.bEggOptchar.val)
			txt.write(MakeSettings(self.bFile.val, selection, self.anims, tools, self.bOptParms.val, self.bMExtraction.val, self.bAnimOnly.val))
		self.buttonFuncs[8] = fSaveSettings

		if self.anims is None:
					name = 'anim'
					fps = Blender.Scene.GetCurrent().getRenderingContext().framesPerSec()
					start = Blender.Get('staframe')
					end = Blender.Get('endframe')
					self.anims = [[name, fps, start, end]]
		self.animationRows = [Gui.AnimationRow(*a) for a in self.anims]

		Draw.Register(self.draw, self.event, self.button)
		Draw.Draw()
		self.analyzeSelection()

		self.loadState() # Will only overwrite the above setting of animationRows if there is something to overwrite with.		

	def draw(self):
		ui = Theme.Get()[0].get(-1)
		col = ui.setting
		col = map(lambda x: (x+40)/255.0, col[0:3])
		area = Window.GetAreaSize()
		w = area[0]; wPad = int(w*0.03)
		h = area[1]; hPad = int(h*0.03)
		#Draw Header
		BGL.glColor3f(*col)
		BGL.glRectf(wPad, h-hPad-10, w-wPad, h-hPad-40)
		if self.logo:
			BGL.glEnable( BGL.GL_BLEND ) # Only needed for alpha blending images with background.
			BGL.glBlendFunc(BGL.GL_SRC_ALPHA, BGL.GL_ONE_MINUS_SRC_ALPHA)
			Draw.Image(self.logo, wPad, h-hPad-90)
			BGL.glDisable( BGL.GL_BLEND )
		BGL.glColor3f(0.0,0.0,0.0)
		BGL.glRasterPos2i(w-wPad-self.vMsgW, h-hPad-30)
		Draw.Text(self.vMsg)

		BGL.glColor3f(0.3, 0.3, 0.3)
		BGL.glRectf(wPad, h-90, w-wPad, h-110)
		BGL.glColor3f(1.0,1.0,1.0)
		BGL.glRasterPos2i(wPad, h-105)
		if self.page is 0:
			Draw.Text('Status: ')
			if self.error:
				BGL.glColor3f(0.9, 0.0, 0.0)
				BGL.glRasterPos2i(wPad+40, h-105)
			Draw.Text(self.statusMsg)
			if self.success:
				if self.error:
					var = 'with errors.'
				else:
					var = 'successfully!'
				Draw.Text(' ...Exported %s  Time:%.3f sec'%(var,self.time))
			BGL.glColor3f(1.0, 1.0, 1.0)
			self.bReport.name = 'View Report Page'
			self.bReport.update([w-wPad-210, h-110, 110, 20])
			self.bUpdate.update([w-wPad-100, h-110, 100, 20])
			self.bFile.update([wPad+5, h-140, w-2*wPad-90, 20])
			self.bSFile.update([w-wPad-85, h-140, 80, 20])
			if len(self.meshes['armature']) > 0:
				BGL.glColor3f(0.5, 0.5, 0.5)
				BGL.glRectf(wPad, h-150, w-wPad, h-170)
				BGL.glColor3f(1.0,1.0,1.0)
				BGL.glRasterPos2i(wPad, h-165)
				Draw.Text('Animations:')
				top = h-200
				for i, r in enumerate(self.animationRows):
					r.update(i, w, wPad, top)
					top -= 22
				self.bAddAnim.update([wPad+20, top-1, 100, 20])
				if self.srb:
					self.bMExtraction.update([w-wPad-500, top-1, 110, 20])
				self.bDoTangents.update([w-wPad-390, top-1, 150, 20])
				self.bSingleFile.update([w-wPad-230, top-1, 100, 20])
				self.bAnimOnly.update([w-wPad-120, top-1, 100, 20])
			else:
                                self.bDoTangents.update([w-170, h-201, 150, 20])
			self.bQuit.update([wPad, hPad, 40, 30])
			self.bHelp.update([wPad+45, hPad, 40, 30])
			BGL.glColor3f(1.0,1.0,1.0)
			BGL.glRasterPos2i(wPad+95, hPad+10)
			Draw.Text('Call:')
			self.bEgg2Bam.update([wPad+130, hPad+5, 60, 20])
			self.bPview.update([wPad+190, hPad+5, 40, 20])
			if len(self.meshes['armature']) > 0:
				self.bEggOptchar.update([wPad+230, hPad+5, 80, 20])
				self.bOptParms.update([wPad+310, hPad+5, w-2*wPad-400, 20])
			if MakeSettings:
				self.bExport.update([w-wPad-85, hPad, 40, 30])
				self.bSaveSettings.update([w-wPad-40, hPad, 40, 30])
			else:
				self.bExport.update([w-wPad-85, hPad, 40, 30])

		elif self.page is 1:
			BGL.glRasterPos2i(wPad+5, h-105)
			Draw.Text('Report Page')
			self.bReport.name = 'Go Back'
			self.bReport.update([w-wPad-80, h-110, 80, 20])
			top = 125
			for m in self.messages:
				if m.severity == Message.Error:
					BGL.glColor3f(0.9, 0.0, 0.0)
				elif m.severity == Message.Warning:
					BGL.glColor3f(0.9, 0.45, 0.0)
				else:
					BGL.glColor3f(1.0, 1.0, 1.0)
				BGL.glRectf(wPad+5, h-top+15, w-wPad-5, h-top-5)
				BGL.glColor3f(0.0, 0.0, 0.0)
				BGL.glRasterPos2i(wPad+10, h-top)
				Draw.Text(m.message)
				top += 20

	def event(self, evt, val):
		if evt == Draw.ESCKEY or evt == Draw.QKEY:
                        self.saveState()
			Draw.Exit()
		return

	def button(self, evt):
		if evt >= Gui.animStart:
			i = evt-Gui.animStart
			self.animationRows.pop(i)
		elif evt > 0:
			self.buttonFuncs[evt]()
		Draw.Redraw()

	def saveState(self):
                """Saves the animation list to a property in each animated mesh selected. Stores it as a string for ease of processing by fleshies."""
                data = []
                for row in self.animationRows:
                        data.append((row.bName.val,row.bFps.val,row.bStart.val,row.bEnd.val))

                data = map(lambda x: x[0]+','+str(x[1])+','+str(x[2])+','+str(x[3]),data)
                if len(data)>1:
                        data = reduce(lambda x,y:x+';'+y,data)
                elif len(data)==1:
                        data = data[0]
                else:
                        data = None

                for mesh in self.meshes['armature']:
                        if data:
                                mesh.properties['chicken_ani'] = data
                        else:
                                del mesh.properties['chicken_ani']

        def loadState(self):
                """Loads the animation list as saved from above. Does nothing if theres nothing to load."""
                try:
                        for mesh in self.meshes['armature']:
                                if mesh.properties.has_key('chicken_ani'):
                                        data = mesh.properties['chicken_ani'].split(';')
                                
                                        self.animationRows = []
                                        for d in data:
                                                item = d.rsplit(',',3)
                                                self.animationRows.append(Gui.AnimationRow(item[0],int(item[1]),int(item[2]),int(item[3])))
                                        break
                except Exception, e:
                        print e

class Background(Exporter):
	'''
	This class encapsulates the things involved in user interaction through
	Blender's background mode or through chicken's interface Invoke function
	(when called from other scripts)
	'''
	def __init__(self):
		if FORCE_BACKGROUND:
			print "Entering Chicken Forced Background mode. If you didn't call \
Chicken through the invocation function please unset the \
CHICKEN_FORCE_BACKGROUND environment variable"
		self.init(settings = os.getenv(SETTINGS_KEY), progress_range=os.getenv(RANGE_KEY))

		if FORCE_BACKGROUND:
			if Blender.mode == 'interactive':
				self.reportProgress = DrawProgressBar
			msg_txt = Text.New('messages.chicken')
			os.environ[MESSAGE_KEY] = msg_txt.getName()

		self.analyzeSelection()
		if FORCE_BACKGROUND:
			pre = [self.messages[:]]
		else:
			print "Pre-Export messages:"
			print self.statusMsg
			for m in self.messages:
				print m

		self.doExport()
		if self.success and not FORCE_BACKGROUND:
			print 'Success! Total Export time (including invocations): %.3f seconds'%self.time
			print 'Geometry: %s'%self.filename
			if len(self.anims) > 0:
				anim_names = ' '.join([a[0] for a in self.anims])
				print 'Animations: %s'%anim_names

		if FORCE_BACKGROUND:
			pre.extend([self.messages])
			try:
				import cPickle
				msg_txt.write(cPickle.dumps(tuple(pre), protocol=0))
			except:
				pass
		else:
			print 'Post-Export messages:'
			print self.statusMsg
			for m in self.messages:
				print m
	def reportProgress(self, prog, msg):
		print 'Exporting... %.2f percent done'%(prog*100)

class Material:
	def __init__(self, mat, group):
		self.name = mat.name.replace(' ', '_') # egg files don't like spaces in names
		self.users = 0
		factor = mat.spec/2.0
		self.diff = (mat.R, mat.G, mat.B)
		self.spec = (mat.specR*factor, mat.specG*factor, mat.specB*factor)
		self.shine = mat.hard/4.0
		self.vcol = mat.mode & BMaterial.Modes.VCOL_LIGHT
		self.shadeless = mat.mode & BMaterial.Modes.SHADELESS
		self.textures = []
		# process textures
		
		if group.mesh:
			uvnames = group.mesh.getUVLayerNames()
			if len(uvnames) > 0:
				textures = [] #textures that use UVs
				texgenTextures = [] #here we'll store the MTex of textures that don't use UVs
				texcount = 0
				for mtex in mat.getTextures():
					if type(mtex) == Types.MTexType and mtex.tex.getType() == 'Image':
						if mtex.texco == BTexture.TexCo.UV:
							if mtex.uvlayer in uvnames: uv = mtex.uvlayer													
							else: uv = uvnames[0] # if the UVLayer isn't valid for this mesh use the first
							list = textures
						else:
							uv = None
							list = texgenTextures
							
						if mtex.mtCol == 1 or mtex.mtNor == 1:
							if mtex.mtNor == 1 and not mat.mode & BMaterial.Modes.NMAP_TS:
								exporter.messages.append(Message('Material %s is not using tangent space normal maps. Appearance will differ from Blender.'%self.name, Message.Warning))
							list.append([mtex,None,uv])
							texcount += 1
						elif mtex.mtAlpha == 1:
							appended = False
							for texSpec in reversed(list):
								if texSpec[0].mtCol == 1 and texSpec[2] == uv:
									appended = True
									texSpec[1] = mtex #place as alpha texture of last color texture on this UVlayer
									break
							if not appended:
								list.append([mtex, None, uv]) #place as its own texture
						
				for i,texSpec in enumerate(textures):
					filename = texSpec[0].tex.getImage().getFilename()
					name = self.name
					if texcount > 1: 					
						name += str(i)
					pTex = Texture(name, filename, texSpec[2], texSpec[0])
					TEXTURES[name] = pTex
					self.textures.append(name)
					if texSpec[1]:
						aFilename = None
						aImage = texSpec[1].tex.getImage()
						if texSpec[1] is not texSpec[0] and aImage:
							aFilename = aImage.getFilename()
						pTex.setAlphaChannel(aFilename)
				
				#TODO: process texgenTextures
	def getRef(self):
		for tex in self.textures:
			TEXTURES[tex].users += 1
		return self.textures
	def __repr__(self):
		return '[Material %s, users:%i]'%(self.name, self.users)
	def write(self, buffer):
		ln = ['<Material> %s {'%self.name,
			'  <Scalar> diffr {%s}\n  <Scalar> diffg {%s}\n  <Scalar> diffb {%s}'%self.diff,
			'  <Scalar> specr {%s}\n  <Scalar> specg {%s}\n  <Scalar> specb {%s}'%self.spec,
			'  <Scalar> shininess {%s}'%self.shine,
			'}']
		buffer.extend(ln)

class Texture:
	def __init__(self, name, filename, uvname=None, mtex=None):
		self.name = name.replace(' ', '_')
		self.users = 0
		self.filename = self.convertFileNameToPanda(filename)
		self.hasAlpha = mtex and mtex.mtAlpha == 1
		self.alphaFileName = None
		self.uvname = uvname
		self.envtype = 'modulate'
		if mtex and mtex.mtNor ==  1 and uvname:
			if not mtex.tex.imageFlags & BTexture.ImageFlags.NORMALMAP:
				exporter.messages.append(Message("Blender Texture %s is mapped to normals, but it's not flagged as a normal map"%mtex.tex.name, Message.Warning))
			exporter.normalMappedUVs.add(uvname)
			self.envtype = 'normal'
	def setAlphaChannel(self, alphaFileName = None):
		self.hasAlpha = True
		if alphaFileName:
			self.alphaFileName = self.convertFileNameToPanda(alphaFileName)
	def convertFileNameToPanda(self, filename):
		path =  filename.replace('//', './').replace('\\', '/')
		if os.name == 'nt' and path.find(':') != -1:
			path = '/'+ path[0].lower() + path[2:]
		return path
	def __repr__(self):
		return '[Texture %s, users:%i, fn:%s, a:%s]'%(self.name, self.users, self.filename, str(self.hasAlpha))
	def write(self, buffer):
		ln = ['<Texture> %s {'%self.name,
			'  "%s"'%self.filename]
		if self.uvname:
			ln.append('  <Scalar> uv-name { %s }'%self.uvname)
		if self.hasAlpha:
			if self.alphaFileName:
				ln.append('  <Scalar> alpha-file { "%s" }'%self.alphaFileName)
			ln.append('  <Scalar> alpha { blend_no_occlude }')
			ln.append('  <Scalar> format { rgba }')
			ln.append('  <Scalar> draw_order { 10000 }')
		ln.append('  <Scalar> envtype { %s }'%self.envtype)
		ln.append('}')
		buffer.extend(ln)

class Group:
	def __init__(self, name = '', mesh = None, anim = False, addTangents = False):
		self.children = []
		self.polys = []
		self.verts = []
		self.bverts2mverts = None
		self.anim = anim
		self.tags = {}
		self.objectType = None
		if not mesh:
			self.name = name.replace(' ', '_') # egg files don't like spaces in names
			self.transform = None
		else:
			self.name = mesh.name.replace(' ', '_')
			# grab and interpret properties
			props = mesh.getAllProperties()
			for p in props:
				name = p.getName()
				data = str(p.getData())
				self.tags[name] = data
			if 'ObjectType' in self.tags:
				self.objectType = self.tags.pop('ObjectType')
			if 'chicken_ani' in self.tags:
				self.objectType = self.tags.pop('chicken_ani')				

			self.transform = mesh.matrix

			if mesh.getType() == 'Empty': #empties get tags and the transform
				if not self.objectType:
					self.objectType = 'model' #prevent flattening of empties unless otherwise specified
				return

			scn = Blender.Scene.GetCurrent()
			self.tmesh = Mesh.New()
			self.tmesh.getFromObject(mesh,anim)
			self.tmesh.transform(mesh.matrix) #make a transformed mesh copy to speed up verts
			tempobj = scn.objects.new(self.tmesh)
			print tempobj
			
			self.mesh = mesh.getData(mesh=1)

			if addTangents and (len(self.tmesh.getUVLayerNames())!=0):
                                tangents = []
                                for uvLayer in self.tmesh.getUVLayerNames():
                                        self.tmesh.activeUVLayer = uvLayer
                                        tangents.append(self.tmesh.getTangents())
                        else:
                                tangents = None

			hasUV, hasVCol = self.mesh.faceUV, self.mesh.vertexColors
			mats = self.mesh.materials
			if anim:
				self.transform = None
				self.bverts2mverts = {}
			for mat in mats: # process materials that haven't been processed yet
				if mat and mat.name not in MATERIALS:
					MATERIALS[mat.name] = Material(mat, self)
			if len(mats) <= 1:
				for i,face in enumerate(self.tmesh.faces):
					self.polys.append(Polygon(face, i, self, hasUV, hasVCol, mats, tangents))
			else: # group by material indices if there's more than one
				self.children = [Group(self.name+str(i+1)) for i in xrange(len(mats))]
				for i,face in enumerate(self.tmesh.faces):
					p = Polygon(face, i, self, hasUV, hasVCol, mats, tangents)
					self.children[face.mat].polys.append(p)
				# filter those unused
				self.children = filter(lambda c: len(c.polys) > 0, self.children)
			#remove transformed mesh
			scn.objects.unlink(tempobj)
			del self.tmesh
	def __repr__(self, padding = ''):
		ln = [padding+'[Group %s, polys:%i, verts:%i]'%(self.name, len(self.polys), len(self.verts))]
		padding += ' '
		ln.extend([c.__repr__(padding) for c in self.children])
		return '\n'.join(ln)
	def write(self, buffer, padding = ''):
		buffer.append('%s<Group> %s {'%(padding, self.name))
		self.writeInner(buffer, padding)
		buffer.append('%s}'%(padding))
	def writeInner(self, buffer, padding = ''): # separated for Character reuse
		self.writeTransform(buffer, padding)
		padding2 = padding+'  '
		for k,v in self.tags.iteritems():
			buffer.append('%s<Tag> %s { %s }'%(padding2, k, v))
		if self.objectType:
			buffer.append('%s<ObjectType> { %s }'%(padding2, self.objectType))
		if len(self.verts) > 0:
			buffer.append('%s<VertexPool> %s {'%(padding2, self.name))
			padding3 = padding2+'  '
			for v in self.verts:
				v.write(buffer, padding3)
			buffer.append('%s}'%(padding2))
		for p in self.polys:
			p.write(buffer, padding2)
		for c in self.children:
			c.write(buffer, padding2)
	def writeTransform(self, buffer, padding): # separated for Joint reuse
		if self.transform and self.transform != IDENTITY:
			m = self.transform
			buffer.append('%s  <Transform> {'%padding)
			buffer.append('%s    <Matrix4> {'%padding)
			buffer.append('%s      %f %f %f %f'%(padding, m[0][0], m[0][1], m[0][2], m[0][3]))
			buffer.append('%s      %f %f %f %f'%(padding, m[1][0], m[1][1], m[1][2], m[1][3]))
			buffer.append('%s      %f %f %f %f'%(padding, m[2][0], m[2][1], m[2][2], m[2][3]))
			buffer.append('%s      %f %f %f %f'%(padding, m[3][0], m[3][1], m[3][2], m[3][3]))
			buffer.append('%s    }'%padding)
			buffer.append('%s  }'%padding)


class Character(Group):
	def __init__(self, armature, meshes, srb, animOnly = False, addTangents = False):
		'''
		Constructor
		@param armature: The Armature object that defines this character
		@param meshes: A list of Mesh objects that define the geometry for this character
		@param animOnly: Flag that indicates whether this character is being generated solely for exporting animation
		'''
		Group.__init__(self, armature.name,addTangents = addTangents)
		self.armature = armature
		arm_mat = armature.matrix
		self.mat = arm_mat
		if not animOnly:
			self.children = [Group(mesh=m, anim = True) for m in meshes]
		arm_data = armature.getData()
		bones = arm_data.bones.values()
		self.bones = arm_data.bones
		self.boneNames = arm_data.bones.keys()
		if srb:
			srbIndex = self.boneNames.index('SyntheticRootBone') # get the index where I'll find the bone and name
			self.boneNames.pop(srbIndex) # pop the name
			srbBone = bones.pop(srbIndex) # pop the bone so we don't treat it like the other joints
			self.srb = Joint(srbBone, arm_mat, animOnly) # make the joint and store it separately
		else: self.srb = None
		#begin reconstruction of hierarchy
		parents = map(lambda x: x.parent, bones)
		jointList = [Joint(bone, arm_mat, animOnly) for bone in bones]
		joints = dict([(j.origName, j) for j in jointList])
		roots = []
		self.roots = roots
		for i, p in enumerate(parents):
				if p:
					joints[p.name].children.append(jointList[i])
				else:
					roots.append(jointList[i])
		#end
		#begin assigning weight references
		if not animOnly:
			for c in self.children:
				cmesh = c.mesh
				cname = c.name
				mapping = c.bverts2mverts
				for i in xrange(len(cmesh.verts)):
					inf = cmesh.getVertexInfluences(i)
##					if len(inf) > 4: print 'trouble' # maybe do something about that
					for w in inf:
						try:
							joint = joints[w[0]]
							key = (w[1], cname)
							indices = joint.weights.get(key, None)
							if not indices:
								indices = []
								joint.weights[key] = indices
							indices.extend(mapping[i])
						except KeyError: pass # ignore vertex groups that aren't for bones
		#end

	def __repr__(self, padding = ''):
		ln = [padding+'[Character %s, polys:%i, verts:%i]'%(self.name, len(self.polys), len(self.verts))]
		padding += ' '
		ln.extend([c.__repr__(padding) for c in self.children])
		ln.extend([j.__repr__(padding) for j in self.roots])
		return '\n'.join(ln)
	def write(self, buffer, padding = ''):
		buffer.append('%s<Group> %s {'%(padding, self.name))
		buffer.append('%s  <DART> { 1 }'%padding)
		self.writeInner(buffer, padding) # separated so the Character class can reuse this part
		padding2 = padding + '  '
		for j in self.roots:
			j.write(buffer,padding2)
		if self.srb:
			self.srb.write(buffer, padding2)
		buffer.append('%s}'%(padding))
	def writeAnims(self, charfilename, anims, out_file=None):
		'''
		Writes animations specified in a list into separate files whose filenames are based on the one passed in
		@param charfilename: The filename to which the character was exported
		@param anims: List of animation specifications, which should be tuples of the form (name, fps, startframe, endframe)
		'''
		if len(anims) == 0:
			return
		framelims = []
		# calculate total range for all anims
		for a in anims:
			framelims.extend([a[2],a[3]])
		minframe = min(framelims)
		maxframe = max(framelims)
		# collect local posematrices for every frame and bone in that range
		bones = self.bones
		boneNames = self.boneNames
		armature = self.armature
		mat = self.mat
		poses = [[]for name in boneNames] # lists of local posematrices for every bone
		if self.srb:
			srb = self.srb
			srbDeltas = []
		else: srb = None
		srbIMatPrev = IDENTITY
		oldframe = Blender.Get('curframe')
		###### adding stuff to get access to the currentFrame
		scn = Blender.Scene.GetCurrent()
		context = scn.getRenderingContext()
		###### done with preparing change frames 
		for f in xrange(minframe,maxframe+1):
			context.currentFrame(f)
			scn.update(1)
			pbones = armature.getPose().bones
			if srb:
				srbMat = Matrix(pbones['SyntheticRootBone'].poseMatrix)
				srbDeltas.append(srbMat * srbIMatPrev)
				srbIMatPrev = srbMat.invert() # if there is no srb this will always be just identity
			for i, name in enumerate(boneNames):
				pbone = pbones[name]
				parent = bones[name].parent
				if parent:
					imat = (pbones[parent.name].poseMatrix*mat*srbIMatPrev).invert()
					poses[i].append(pbone.poseMatrix*mat*srbIMatPrev * imat)
				else:
					poses[i].append(pbone.poseMatrix*mat*srbIMatPrev)
		Blender.Set('curframe', oldframe)
		poses = dict([(name, poses[i]) for i, name in enumerate(boneNames)]) # convert poses to a dictionary like this (bonename:[posematrices])
		if srb:
			poses['SyntheticRootBone'] = srbDeltas
		# begin export of animations
		for a in anims:
			if a[2] > a[3]:
				r = range(a[2]-minframe, a[3]-minframe-1, -1)
			else:
				r = range(a[2]-minframe, a[3]-minframe+1)
			if out_file is None:
                                buffer = INITIAL[:]
                        else:
                                buffer = ["",]
			buffer.append('<Table> {')
			if out_file is None: # If files are seperate it must be the name of the group it applies to, else the name of the animation.
                                buffer.append('  <Bundle> %s {'%self.name)
                        else:
                                buffer.append('  <Bundle> %s {'%a[0])
			buffer.append('    <Table> "<skeleton>" {')
			padding = '      '
			for j in self.roots:
				j.writeAnim(a[1], r, poses, buffer, padding)
			if srb:
				self.srb.writeAnim(a[1], r, poses, buffer, padding)
			buffer.append('    }')
			buffer.append('  }')
			buffer.append('}')
			if out_file is None:
                                fn = '%s-%s'%(charfilename, a[0])
                                out_file = open(makename(fn, '.egg'),"w")
                                out_file.write('\n'.join(buffer))
                                out_file.close()
                                out_file = None
                        else:
                                out_file.write('\n'.join(buffer))
class Vertex:
	def __init__(self, vert, index, color=None, normal=False, multiuv=None, tangents=None, faceNormal=None):
		self.index = index
		self.coords = vert.co
		self.color = color
		if normal: self.normal = vert.no
		else: self.normal = None
		if multiuv:
                        if tangents:
                                self.doTangent = True
                                if not self.normal: # We need a normal to cross with for flat shaded faces to get the bi-tangent.
                                        self.faceNormal = faceNormal
                                mesh, face_i, vert_i = multiuv
                                def getMultiTangentUV(name,tans):
                                        mesh.activeUVLayer = name
                                        uv = mesh.faces[face_i].uv[vert_i]
                                        tangent = tans[face_i][vert_i]
                                        return (name, uv, tangent)
                                self.uv = map(getMultiTangentUV, mesh.getUVLayerNames(), tangents)
                        else:
                                self.doTangent = False
                                mesh, face_i, vert_i = multiuv
                                def getMultiUV(name):
                                        mesh.activeUVLayer = name
                                        uv = mesh.faces[face_i].uv[vert_i]
                                        return (name, uv)
                                self.uv = map(getMultiUV, mesh.getUVLayerNames())
		else:
			self.uv = None
	def write(self, buffer, padding = ''):
		padding2 = padding+'  '
		coord_str = ' '.join(map(str, self.coords)) #this way it supports arbitrary number of coords, for NURBS
		ln = ['%s<Vertex> %i {'%(padding, self.index),
				'%s%s'%(padding2, coord_str)]
		if self.uv:
                        if self.doTangent==False:
                                for name, uv in self.uv:
                                        ln.append('%s<UV> %s { %f %f }'%(padding2, name, uv[0], uv[1]))
                        else:
                                for name, uv, tangent in self.uv:
                                        if self.normal:
                                                bin = -CrossVecs(self.normal,tangent)
                                        else:
                                                bin = -CrossVecs(self.faceNormal,tangent)
                                        
                                        ln.append('%s<UV> %s {'%(padding2, name))
                                        ln.append('%s  %f %f'%(padding2,uv[0],uv[1]))
                                        ln.append('%s  <Tangent> { %f %f %f }'%(padding2,tangent[0],tangent[1],tangent[2]))
                                        ln.append('%s  <Binormal> { %f %f %f }'%(padding2,bin[0],bin[1],bin[2]))
                                        ln.append('%s}'%(padding2,))
		if self.normal:
			ln.append('%s<Normal> { %f %f %f }'%(padding2, self.normal[0], self.normal[1], self.normal[2]))
		if self.color:
			col = self.color
			col = [col.r, col.g, col.b, col.a]
			col = ' '.join(map(lambda x: str(x/255.0), col))
			ln.append('%s<RGBA> { %s }'%(padding2, col))
		ln.append(padding+'}')
		buffer.extend(ln)

class Polygon:
	def __init__(self, face, faceindex, group, hasUV, hasVCol, materials, tangents):
		'''
		Constructor
		@param face: The face on which to base this polygon
		@param group: The group that contains the vertex pool for this face
		@param hasUV: indicates whether the mesh has face UV coordinates
		@param hasVCol: indicates whether the mesh has vertex colors
		@param materials: list of the mesh's materials
		@param tangents: list of results of the mesh.getTangents() function for each uv layer, in the order of mesh.getUVLayerNames().
		'''
		self.matRef = None
		self.texRef = []
		mat = None
		if len(materials) > 0:
			try: mat = MATERIALS[materials[face.mat].name]
			except AttributeError: pass
			except KeyError: pass
			else:
				if not mat.shadeless:
					self.matRef = mat.name # store material reference
					mat.users += 1
				self.texRef.extend(mat.getRef()) # store texture reference
		if not mat or len(mat.textures) is 0: # if we don't have a material or it has no textures, use image textures
			for uvLayer in group.tmesh.getUVLayerNames():
				group.tmesh.activeUVLayer = uvLayer
				if hasUV and face.mode & Mesh.FaceModes.TEX and face.image and face.image.name not in self.texRef:
					# create texture for face image if it doesn't exist
					if face.image.name not in TEXTURES:
						tex = Texture(face.image.name, face.image.filename, uvLayer)
						TEXTURES[face.image.name] = tex
					else:
						tex = TEXTURES[face.image.name]
					tex.users += 1
					self.texRef.append(tex.name)
		self.ref = group.name # store group name for vertex pool reference
		self.bface = hasUV and face.mode & Mesh.FaceModes.TWOSIDE
		if face.smooth: self.normal = None
		else: self.normal = face.no
		
		
		i = len(group.verts)
		self.indices = range(i, i+len(face.v))
		smooth = face.smooth
		mapping = group.bverts2mverts
		for i in xrange(len(face.v)):
			params = {'vert':face.v[i], 'index':self.indices[i], 'tangents':tangents, 'faceNormal':self.normal}
			if hasUV:
				params['multiuv'] = (group.tmesh, faceindex, i)
			if hasVCol or (mat and mat.vcol): params['color'] = face.col[i]
			if smooth: params['normal'] = True
			mvert = Vertex(**params)
			group.verts.append(mvert)
			if mapping is not None:
				bvert = mapping.get(face.v[i].index, None)
				if not bvert:
					bvert = []
					mapping[face.v[i].index] = bvert
				bvert.append(self.indices[i])
		
	def write(self, buffer, padding = ''):
		padding2 = padding+'  '
		ln = [padding+'<Polygon> {']
		for tex in self.texRef:
			ln.append('%s<TRef> { %s }'%(padding2, tex))
		if self.matRef:
			ln.append('%s<MRef> { %s }'%(padding2, self.matRef))
		if self.bface:
			ln.append('%s<BFace> { 1 }'%(padding2))
		if self.normal:
			ln.append('%s<Normal> { %f %f %f }'%(padding2, self.normal[0], self.normal[1], self.normal[2]))
		ln.append('%s<VertexRef> { %s <Ref> { %s } }'%(padding2, ' '.join(map(str, self.indices)), self.ref))
		ln.append(padding+'}')
		buffer.extend(ln)

class Joint(Group):
	def __init__(self, bone, mat, animOnly, isSrb = False):
		'''
		Constructor
		@param bone: The bone on which to base this joint
		@param mat: The containing Armature's world transform
		@param animOnly: A flag that indicates whether this bone is being generated solely for exporting animation
		'''
		Group.__init__(self,bone.name)
		self.origName = bone.name # original name with possible spaces, used for looking up pose data
		del self.bverts2mverts
		if not animOnly:
			if not bone.parent:
				self.transform = bone.matrix['ARMATURESPACE'] * mat
			else:
				imat = bone.parent.matrix['ARMATURESPACE'] * mat
				imat.invert()
				self.transform = bone.matrix['ARMATURESPACE'] * mat * imat
			self.weights = {} #dictionary of weights [(weight, poolname):[indices]]
	def __repr__(self, padding = ''):
		ln = [padding+'[Joint %s]'%(self.name)]
		padding += ' '
		ln.extend([c.__repr__(padding) for c in self.children])
		return '\n'.join(ln)
	def write(self, buffer, padding = ''):
		buffer.append('%s<Joint> %s {'%(padding, self.name))
		self.writeTransform(buffer, padding)
		for k, v in self.weights.iteritems():
			buffer.append('%s  <VertexRef> {'%padding)
			buffer.append('%s    %s'%(padding, ' '.join(map(str,v))))
			buffer.append('%s    <Scalar> membership { %f }'%(padding, k[0]))
			buffer.append('%s    <Ref> { %s }'%(padding, k[1]))
			buffer.append('%s  }'%padding)
		padding2 = padding + '  '
		for c in self.children:
			c.write(buffer, padding2)
		buffer.append('%s}'%(padding))
	def writeAnim(self, fps, aRange, poses, buffer, padding):
		buffer.append('%s<Table> %s {'%(padding, self.name))
		buffer.append('%s  <Xfm$Anim> xform {'%padding)
		buffer.append('%s    <Scalar> fps { %i }'%(padding,fps))
		buffer.append('%s    <Scalar> order { sprht }'%padding)
		buffer.append('%s    <Scalar> contents { ijkprhxyz }'%padding)
		buffer.append('%s    <V> {'%padding)
		mats = poses[self.origName]
		for f in aRange:
			mat = mats[f]
			vals = list(mat.scalePart())
			vals.extend(list(mat.toEuler()))
			vals.extend(list(mat.translationPart()))
			vals = map(str, vals) #convert all to strings
			buffer.append('%s      %s'%(padding, ' '.join(vals)))
		buffer.append('%s    }'%padding)
		buffer.append('%s  }'%padding)
		padding2 = padding + '  '
		for c in self.children:
			c.writeAnim(fps, aRange, poses, buffer, padding2)
		buffer.append('%s}'%(padding))

class NurbsCurve(Group):
	def __init__(self, name, curnurb, obj):
		Group.__init__(self, name)
		knots = curnurb.knotsU
		self.order = len(knots)-curnurb.points
		self.knots = ' '.join(map(str, knots))
		self.indices = ' '.join(map(str, range(curnurb.points)))
		class DummyVert(Vertex):
			def __init__(self, nurbpoint, index):
				self.coords = nurbpoint[0:4]
				self.index = index
				self.uv = self.color = self.normal = None
		self.verts = []
		for i, p in enumerate(curnurb):
			self.verts.append(DummyVert(p, i))

	def write(self, buffer):
		buffer.append('<Group> %s {'%self.name)
		self.writeInner(buffer)
		buffer.append('  <NURBSCurve> {')
		buffer.append('    <Order> { %i }'%self.order)
		buffer.append('    <Knots> { %s }'%self.knots)
		buffer.append('    <VertexRef> { %s  <Ref> { %s }}'%(self.indices, self.name))
		buffer.append('  }')
		buffer.append('}')

### Runtime entry point
if not INTERFACE_ERROR:
	FORCE_BACKGROUND = os.getenv(KEY)
	if Blender.mode == 'background' or FORCE_BACKGROUND:
		if PATH_ERROR:
			print "There is an error in Chicken's configuration. Please run from GUI to access the configuration script"
		else:
			exporter = Background()
	elif not PATH_ERROR:
		import webbrowser
		exporter = Gui()
	else:
		Blender.Run(os.path.join(Blender.Get('datadir'),'chicken/chicken_config.py'))
